
#include <stdio.h>
#include <list>

#include "Globals.h"

#include "Context.h"
#include "Document.h"
#include "Item.h"
#include "Parser.h"
#include "BuildDoc.h"

static Document *g_pDocument;
Context g_CurrentContext;
list<ListNode> g_TagNodeStack;
int g_nTotalItems = 0;

void HandleText (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  DocItemHandle newItem( new DocTextItem(pcSource, nLineno));
  AddItem(g_pDocument, newItem);
}

void HandleWhitespace (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  DocItemHandle newItem( new DocWhitespaceItem(pcSource, nLineno));
  AddItem(g_pDocument, newItem);
}

void HandleUnknownTag (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  char pcTemp[1024];
  sprintf(pcTemp, "unknown tag %s", pcSource);
  MyError(pcTemp, nLineno, 0);
}

void HandleTag (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  ItemType itemType = (ItemType) (int) pArg;

  DocItemHandle newItem( new DocTagItem(itemType, bStart, nLineno) );

  ListNode otherItemNode;
  if (!bStart) {                        // an end tag
//    assert(g_TagNodeStack.size() > 0);
    assert(g_TagNodeStack.begin() != g_TagNodeStack.end());
    otherItemNode = g_TagNodeStack.front();
    assert( (*otherItemNode)->m_ItemType == itemType);  // should be the same item type
    g_TagNodeStack.pop_front();
    newItem->m_pOtherItemNode = new ListNode(otherItemNode);
  }

  ListNode lastItem = AddItem(g_pDocument, newItem);

  if (!bStart) {                        // an end tag
    (*otherItemNode)->m_pOtherItemNode = new ListNode(lastItem);
  } else {                              // a start tag
    g_TagNodeStack.push_front(lastItem);
  }

}

ListNode AddItem (Document *pDocument, DocItemHandle newItem) {

  g_nTotalItems++;
  if ((g_nTotalItems % 1000) == 0) {
//    fprintf(stderr, "%d items parsed\n", g_nTotalItems);
  }

//  if (pDocument->size() > 0) {          // not the first item
  if (pDocument->begin() != pDocument->end()) {
    newItem->m_Context = g_CurrentContext;
  } else {                              // first item
    // do nothing, default property is root
  }
  g_CurrentContext = Context::TransformContext(&g_CurrentContext, newItem);
  return g_pDocument->insert(g_pDocument->end(), newItem);
}

void BuildDoc(Document *pDocument, const char *pSource, int nSize) {
  Parser *pParser = new Parser(HandleText, HandleWhitespace, HandleUnknownTag);
  char pcTemp[1024];
  for (int i=0; i<(int) ITEM_NUM_ENUMS - (int) FIRST_TAG_ITEM; i++) {
    int index = i + (int) FIRST_TAG_ITEM;
    strcpy(pcTemp, g_TagBeginString[index]);
    pcTemp[strlen(g_TagBeginString[index]) - 1] = '\0';
    pParser->registerCallback(pcTemp + 1, HandleTag, (void*) index);
  }
  g_pDocument = pDocument;
  g_CurrentContext = Context();
  pParser->parse(pSource, nSize);

  delete pParser;
}
