
#include "Document.h"
#include "Item.h"

int Document::docSize() {
  iterator item;
  int nResult = 0;
  for (item = begin(); item != end(); item++) {
    nResult += (*item)->size();
  }
  return nResult;
}

int Document::write(char *pDest) {
  iterator item;
  int nResult = 0;
  for (item = begin(); item != end(); item++) {
    nResult += (*item)->write(pDest + nResult);
  }
  return nResult;
}
/*
bool Document::CompareOutput(Document const &doc1, Document const &doc2, DocItemHandle &outItem1, DocItemHandle &outItem2, string &sOutReason) {
  // go through every text and whitespace item and compare the property record

  const_iterator item1, item2;
  item1 = doc1.begin();
  item2 = doc2.begin();

  while (1) {
    // find the next text or whitespace item
    while (item1 != doc1.end() && ! ((*item1)->IsText() || (*item1)->IsWhitespace()) ) {
      item1++;
    }
    while (item2 != doc2.end() && ! ((*item2)->IsText() || (*item2)->IsWhitespace()) ) {
      item2++;
    }

    if (item1 == doc1.end() || item2 == doc2.end()) {
      break;
    }

    outItem1 = (*item1);    
    outItem2 = (*item2);    


    char pcTemp[1024];
    int c = 0;
    c += (*item1)->write(pcTemp);
    sprintf(pcTemp + c, " : ");
    c += (*item2)->write(pcTemp);
    sprintf(pcTemp + c, "\n");    
    printf("%s", pcTemp);

    if ((*item1)->m_ItemType != (*item2)->m_ItemType) {
      sOutReason = "type mismatch";
      return false;
    }
    
    if ((*item1)->IsText()) {
      if (! Context::CompareTextOutput(*item1, *item2)) {
        sOutReason = "text output mismatch";
        return false;
      }
    } else if ((*item1)->IsWhitespace()) {
      if (! Context::CompareWhitespaceOutput(*item1, *item2)) {
        sOutReason = "whitespace output mismatch";
        return false;
      }
    } else {
      assert(false);
    }
    
    item1++;
    item2++;
  }

  // at least one list has ended, find if there are any more
  if (item1 != doc1.end()) {
    for (; item1 != doc1.end(); item1++) {
      if ((*item1)->IsText() || (*item1)->IsWhitespace()) {
        outItem1 = (*item1);    
        sOutReason = "list size mismatch";
        return false;
      }
    }
  }
  if (item2 != doc2.end()) {
    for (; item2 != doc2.end(); item2++) {
      if ((*item2)->IsText() || (*item2)->IsWhitespace()) {
        outItem2 = (*item2);    
        sOutReason = "list size mismatch";
        return false;
      }
    }
  }
  
  return true;
}
*/
void Document::RemoveTagPair(iterator item) {
  assert ((*item)->IsTag());
  assert ((*item)->m_pOtherItemNode != NULL);
  assert ( *((**((*item)->m_pOtherItemNode))->m_pOtherItemNode) == item );

  erase( *((*item)->m_pOtherItemNode));
  erase(item);
}

void Document::CollectString(const_iterator iBegin, const_iterator iEnd, bool bIgnoreTags, string &outString) {
  for (; iBegin != iEnd; iBegin++) {
    if (bIgnoreTags) {
      if (!(*iBegin)->IsTag()) {
        outString += (*iBegin)->m_sText;
      }
    }
  }
}
  
bool Document::CompareOutput(Document const &doc1, Document const &doc2, 
	const_iterator &outBegin1, const_iterator &outEnd1, const_iterator &outBegin2, const_iterator &outEnd2)
{
  const_iterator item1, item2, begin1, end1, begin2, end2;

  item1 = doc1.begin();
  item2 = doc2.begin();

  while (1) {
    // find the next sequence of equivalent text or whitespace
    FindNextSequence(item1, doc1.end(), begin1, end1);    
    FindNextSequence(item2, doc2.end(), begin2, end2);

    outBegin1 = begin1;
    outEnd1 = end1;
    outBegin2 = begin2;
    outEnd2 = end2;
    
    if (!CompareSequence(begin1, end1, begin2, end2)) {
      return false;
    }
    
    bool bEmpty1 = (begin1 == end1);
    bool bEmpty2 = (begin2 == end2);
    if (bEmpty1 && bEmpty2) {
      return true;
    }
    if (bEmpty1 || bEmpty2) {
      return false;
    }
    item1 = end1;
    item2 = end2;    
  }
  return false;
}

void Document::FindNextSequence(const_iterator item1, const_iterator itemEnd, const_iterator &outBegin, const_iterator &outEnd) {
  outBegin = outEnd = itemEnd;
  const_iterator item2;

    while (item1 != itemEnd && (*item1)->IsTag()) {    // 1) find next non-tag item
      item1++;
    }

    if (item1 == itemEnd) {  // reached end
      return;
    }

//    (*item1)->print(stdout);
    
    outBegin = item1;    
    Context::CompareFunc *compareFunc = NULL;
    ItemType itemType = (*item1)->m_ItemType;
    if (itemType == ITEM_TEXT) {
      compareFunc = Context::TextContextEquivalent;
    } else if (itemType == ITEM_WHITESPACE) {
      compareFunc = Context::WhitespaceContextEquivalent;   
    } else {
      assert(false);
    }
  
    // 2) find the next non-tag item that does not have an equivalent context
    while (item1 != itemEnd && ((*item1)->IsTag() ||
           ((*item1)->m_ItemType == itemType && (*compareFunc)((*outBegin)->m_Context, (*item1)->m_Context)))) {
      item1++;
    }
    outEnd = item1;
}
    

bool Document::CompareSequence(const_iterator begin1, const_iterator end1, const_iterator begin2, const_iterator end2) {
  bool bEmpty1 = (begin1 == end1);
  bool bEmpty2 = (begin2 == end2);
  if (bEmpty1 && bEmpty2) {
    return true;
  }
  if (bEmpty1 || bEmpty2) {
    return false;
  }

  assert(!(*begin1)->IsTag() && !(*begin2)->IsTag());

  if ((*begin1)->m_ItemType != (*begin2)->m_ItemType) {
    return false;
  }

  if ((*begin1)->IsText()) {
    return Context::CompareTextOutput(begin1, end1, begin2, end2);
  } else {
    return Context::CompareWhitespaceOutput(begin1, end1, begin2, end2);
  }
}
