
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include "Item.h"

const ItemType FIRST_TAG_ITEM = ITEM_TAG_B;

const int g_TagSize[] = {
	0,	// text
	0,	// whitespace
	3,	// B
	4,	// EM
	3,	// I
	4,	// PL
	3,	// S
	4,	// TT
	3,	// U
	3,	// 0
	3,	// 1
	3,	// 2
	3,	// 3
	3,	// 4
	3,	// 5
	3,	// 6
	3,	// 7
	3,	// 8
	3,	// 9
	3,	// r
	3,	// g
	3,	// b
	3,	// c
	3,	// m
	3,	// y
	3,	// k
	3,	// w
	0,	
};

const char* g_TagBeginString[] = {
	NULL,	// text
	NULL,	// whitespace
	"<B>",	// B
	"<EM>",	// EM
	"<I>",	// I
	"<PL>",	// PL
	"<S>",	// S
	"<TT>",	// TT
	"<U>",	// U
	"<0>",	// 0
	"<1>",	// 1
	"<2>",	// 2
	"<3>",	// 3
	"<4>",	// 4
	"<5>",	// 5
	"<6>",	// 6
	"<7>",	// 7
	"<8>",	// 8
	"<9>",	// 9
	"<r>",	// r
	"<g>",	// g
	"<b>",	// b
	"<c>",	// c
	"<m>",	// m
	"<y>",	// y
	"<k>",	// k
	"<w>",	// w
	NULL,	
};

const char* g_TagEndString[] = {
	NULL,	// text
	NULL,	// whitespace
	"</B>",	// B
	"</EM>",// EM
	"</I>",	// I
	"</PL>",// PL
	"</S>",	// S
	"</TT>",// TT
	"</U>",	// U
	"</0>",	// 0
	"</1>",	// 1
	"</2>",	// 2
	"</3>",	// 3
	"</4>",	// 4
	"</5>",	// 5
	"</6>",	// 6
	"</7>",	// 7
	"</8>",	// 8
	"</9>",	// 9
	"</r>",	// r
	"</g>",	// g
	"</b>",	// b
	"</c>",	// c
	"</m>",	// m
	"</y>",	// y
	"</k>",	// k
	"</w>",	// w
	NULL,	
};

const ItemType g_SizeItemTypes[] = {
	ITEM_TAG_0,
	ITEM_TAG_1,
	ITEM_TAG_2,
	ITEM_TAG_3,
	ITEM_TAG_4,
	ITEM_TAG_5,
	ITEM_TAG_6,
	ITEM_TAG_7,
	ITEM_TAG_8,
	ITEM_TAG_9,
};

const ItemType g_ColorItemTypes[] = {
	ITEM_TAG_r,
	ITEM_TAG_g,
	ITEM_TAG_b,
	ITEM_TAG_c,
	ITEM_TAG_m,
	ITEM_TAG_y,
	ITEM_TAG_k,
	ITEM_TAG_w,
        ITEM_NUM_ENUMS,
};

void AssertArraySizes() {
  assert( (sizeof(g_TagSize) / sizeof(int)) == (int) ITEM_NUM_ENUMS + 1);
  assert( (sizeof(g_TagBeginString) / sizeof(const char*)) == (int) ITEM_NUM_ENUMS + 1);
  assert( (sizeof(g_TagEndString) / sizeof(const char*)) == (int) ITEM_NUM_ENUMS + 1);
}

//////////////////////////////////////////////////////////////
// DocItem

DocItem::DocItem(int nLineno) {
  AssertArraySizes();
  m_ItemType = ITEM_NUM_ENUMS;
  m_pOtherItemNode = NULL;
  m_nLineno = nLineno;
}

DocItem::~DocItem() {
  delete m_pOtherItemNode;
  m_pOtherItemNode = NULL;
}

int DocItem::size() const {
  assert(false);
  return 0;
}

int DocItem::write(char *pDest) const {
  assert(false);
  return 0;
}

//////////////////////////////////////////////////////////////
// DocTextItem

DocTextItem::DocTextItem (const char *pcText, int nLineno)
: DocItem(nLineno)
{
  m_ItemType = ITEM_TEXT;
  m_sText = pcText;
}

int DocTextItem::write(char *pDest) const {
  memcpy(pDest, m_sText.c_str(), m_sText.length());
  return m_sText.length();
}

int DocTextItem::size() const {
  return m_sText.length();
}

//////////////////////////////////////////////////////////////
// DocWhitespaceItem

DocWhitespaceItem::DocWhitespaceItem (const char *pcText, int nLineno) 
: DocTextItem(pcText, nLineno)
{
  m_ItemType = ITEM_WHITESPACE;
//  m_sText = pcText;
}

//////////////////////////////////////////////////////////////
// DocTagItem

DocTagItem::DocTagItem(ItemType itemType, bool bStart, int nLineno)
: DocItem(nLineno)
{
  m_ItemType = itemType;
  m_bStart = bStart;
}

int DocTagItem::size() const {
  assert((int) m_ItemType >= FIRST_TAG_ITEM && (int) m_ItemType < (int) ITEM_NUM_ENUMS);
  return g_TagSize[(int) m_ItemType] + (m_bStart ? 0 : 1);
}

int DocTagItem::write(char *pDest) const {
  memcpy(pDest, (m_bStart ? g_TagBeginString[(int) m_ItemType] : g_TagEndString[(int) m_ItemType]), size());
  return size();
}
  

bool DocItem::CompressWhitespace() {
  bool bResult = false;
  if (m_ItemType == ITEM_WHITESPACE) {
    if (m_sText.length() > 1) {
      bResult = true;
    }
    m_sText = " ";
  }
  return bResult;
}

DocItemHandle DocItem::GetOtherItem() const {
  assert(m_pOtherItemNode != NULL);
  return **m_pOtherItemNode;
}

void DocItem::print (FILE *file) const {
  fprintf(file, "lineno: %d type: %d text: %s start: %d\n", m_nLineno, m_ItemType, m_sText.c_str(), m_bStart);
  m_Context.print(file);
}

void DocItem::SetOtherItemNode (ListNode const item) {
  delete m_pOtherItemNode;
  m_pOtherItemNode = new ListNode(item);
}
