
#ifndef _Item_h
#define _Item_h

#include <string>
#include <assert.h>
#include <stdio.h>

#include "Globals.h"
#include "Context.h"
#include "Document.h"

class DocItem {
public:
  DocItem(int nLineno);
  virtual ~DocItem();

  ItemType m_ItemType;
  ListNode *m_pOtherItemNode;
  string m_sText;
  bool m_bStart;
  int m_nLineno;

  Context m_Context;  // the document property before evaluating this item

  virtual int write(char *pDest) const;
  virtual int size() const;

  void SetOtherItemNode (ListNode item);

  bool IsText() const {
    return m_ItemType == ITEM_TEXT;
  }
  bool IsWhitespace() const {
    return m_ItemType == ITEM_WHITESPACE;
  }
  bool IsTag() const {
    return (m_ItemType != ITEM_TEXT && m_ItemType != ITEM_WHITESPACE && m_ItemType != ITEM_NUM_ENUMS);
  }
  bool IsColor() const {
    return ((int) m_ItemType >= (int) ITEM_TAG_r && (int) m_ItemType <= (int) ITEM_TAG_w);
  }
  bool IsSize() const {
    return ((int) m_ItemType >= (int) ITEM_TAG_0 && (int) m_ItemType <= (int) ITEM_TAG_9);
  }

  DocItemHandle GetOtherItem() const;

  bool CompressWhitespace(); 

  void print (FILE *file) const;
};

class DocTextItem : public DocItem {
public:
  DocTextItem (const char *pcText, int nLineno);

  virtual int write(char *pDest) const;
  virtual int size() const;

};

class DocWhitespaceItem : public DocTextItem {
public:
  DocWhitespaceItem (const char *pcText, int nLineno);
};

class DocTagItem : public DocItem {
public:
  DocTagItem(ItemType itemType, bool bStart, int nLineno);

  virtual int write(char *pDest) const;
  virtual int size() const;
 
};


extern const char* g_TagBeginString[];

#endif //_Item_h
