
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "Optimize.h"
#include "Document.h"
#include "Item.h"
#include "Context.h"


bool CompressWhitespace (Document *pDocument) {
  Document::iterator item;

  bool bCompressed = false;
  for (item = pDocument->begin(); item != pDocument->end(); item++) {
    if ((*item)->IsWhitespace() && (*item)->m_Context.WhitespaceCompressible()) {
      if ((*item)->CompressWhitespace()) {
        bCompressed = true;
      }
    }
  }
  return bCompressed;
}

bool EliminateRedundancies (Document *pDocument) {
  Document::iterator item, otherItem;
  Context context;
  bool bCompressed = false;

  item = pDocument->begin();
  while (item != pDocument->end()) {
    if ((*item)->IsTag()) {
      context = Context::TransformContext( &((*item)->m_Context),  (*item));
      if (context == (*item)->m_Context) {   // tag does not change the context, can eliminate it
        otherItem = *((*item)->m_pOtherItemNode);
 
        item = pDocument->erase(item);
        bCompressed = true;
        if (item == otherItem) {
          item = pDocument->erase(otherItem);
        } else {
          pDocument->erase(otherItem);
        }
        continue;
      }
    }
    item++;
  }
  return bCompressed;    
}

bool EmptyTags (Document *pDocument) {
  Document::iterator item, nextItem;
  Context context;
  bool bCompressed = false;

  item = pDocument->begin();
  while (item != pDocument->end()) {
    if ((*item)->IsTag() && (*item)->m_bStart) {
      nextItem = item;
      nextItem++;
      if ( *((*item)->m_pOtherItemNode) == nextItem) {
        bCompressed = true;
        item = pDocument->erase(item);
        item = pDocument->erase(item);
        continue;
      }
    }
    item++;
  }
  return bCompressed;    
}

bool CombineText (Document *pDocument) {
  Document::iterator item, nextItem;
  bool bCompressed = false;

  item = pDocument->begin();
  ItemType itemType;
  while (item != pDocument->end()) {
    if ((*item)->IsText() || (*item)->IsWhitespace()) {
      itemType = (*item)->m_ItemType;
      nextItem = item;
      nextItem++;
      if (nextItem != pDocument->end()) {
        if ((*nextItem)->m_ItemType == itemType) {
          (*item)->m_sText += (*nextItem)->m_sText;
          bCompressed = true;
          pDocument->erase(nextItem);
        }
      }
    }
    item++;
  }
  return bCompressed;    
}

bool RearrangeWhitespace (Document *pDocument) {
  Document::iterator item, nextItem;
  bool bCompressed = false;

  item = pDocument->begin();
  Document tempDoc;
  while (item != pDocument->end()) {
    if ((*item)->IsTag()) {
      nextItem = item;
      nextItem++;
      if (nextItem != pDocument->end()) {
        if ((*nextItem)->IsWhitespace()) {
          if (Context::WhitespaceContextEquivalent((*item)->m_Context, (*nextItem)->m_Context)) {
            tempDoc.splice(tempDoc.end(), *pDocument, nextItem);   // remove nextItem
            pDocument->splice(item, tempDoc, tempDoc.begin());      // put it back before item
            bCompressed = true;
            continue;
          }
        }
      }
    }
    item++;
  }
  return bCompressed;    
}  

bool EmptyReverseTags (Document *pDocument) {
  Document::iterator item, nextItem, item1, item2;
  Context context;
  bool bCompressed = false;

  item = pDocument->begin();
  while (item != pDocument->end()) {
    if ((*item)->IsTag() && !(*item)->m_bStart) {
      nextItem = item;
      nextItem++;
      if (nextItem != pDocument->end()) {
        if ( (*nextItem)->m_ItemType == (*item)->m_ItemType && (*nextItem)->m_bStart == !(*item)->m_bStart) {
          item1 = *((*item)->m_pOtherItemNode);
          item2 = *((*nextItem)->m_pOtherItemNode);
          bCompressed = true;
          (*item1)->SetOtherItemNode(item2);
          (*item2)->SetOtherItemNode(item1);
          assert((*item1)->m_bStart == true);
          assert((*item2)->m_bStart == false);
          item = pDocument->erase(item);
          item = pDocument->erase(item);
          continue;
        }
      }
    }
    item++;
  }
  return bCompressed;    
}

bool CompressEM (Document *pDocument) {
  Document::iterator item, nextItem, item1, item2;
  Context context;
  bool bCompressed = false;

  item = pDocument->begin();
  while (item != pDocument->end()) {
    if ((*item)->m_ItemType == ITEM_TAG_EM) {
      nextItem = item;
      nextItem++;
      if (nextItem != pDocument->end()) {
        if ((*nextItem)->m_ItemType == ITEM_TAG_EM && (*nextItem)->m_bStart == (*item)->m_bStart) {
          item1 = *((*item)->m_pOtherItemNode);
          item2 = *((*nextItem)->m_pOtherItemNode);
          bCompressed = true;
          (*item1)->SetOtherItemNode(item2);
          (*item2)->SetOtherItemNode(item1);
          assert((*item1)->m_bStart == !(*item)->m_bStart);
          assert((*item2)->m_bStart == !(*nextItem)->m_bStart);
          if ((*item)->m_bStart) {
            (*item2)->m_bStart = true;
          } else { 
            (*item1)->m_bStart = false;
          }
          item = pDocument->erase(item);
          item = pDocument->erase(item);
          continue;
        }
      }
    }
    item++;
  }
  return bCompressed;    
}

bool NestColors (Document *pDocument) {
  ListNodeList colorChain;
  Document::iterator item;
  bool bCompressed = false;

  item = pDocument->begin();
  while (item != pDocument->end()) {
    colorChain.clear();
//    fprintf(stderr, "Calling FindColorChain on item:\n");
//    (*item)->print(stderr);
    if (!FindColorChain(item, pDocument->end(), colorChain)) {
//      fprintf(stderr, "did not find a color chain\n");
      // advance item to after next start color
      while (item != pDocument->end() && (!(*item)->IsColor() || !(*item)->m_bStart)) {
        item++;
      }
      if (item == pDocument->end()) {
        return bCompressed;
      }
      item++;
      continue;
    }
//    fprintf(stderr, "found a color chain\n");
    ListNodeList::iterator cEnd, cStart;
    cEnd = colorChain.begin();
    assert(cEnd != colorChain.end());
    cEnd++;
    assert(cEnd != colorChain.end());
    cStart = colorChain.end();      
    cStart--;
    cStart--;
    assert(cStart != colorChain.begin());
    assert(cStart != cEnd);
    assert(*cStart != *cEnd);
    assert( (*(*cStart))->m_bStart == true);
    assert( (*(*cEnd))->m_bStart == false);
    assert( (*cEnd) != item);
    assert( (*cStart) != item);
    bCompressed = true;
    pDocument->erase(*cEnd);
    pDocument->erase(*cStart);

    cStart = colorChain.begin();
    cEnd = colorChain.end();
    cEnd--;
    assert((**cStart)->m_bStart == true);
    assert((**cEnd)->m_bStart == false);
    (**cStart)->SetOtherItemNode(*cEnd);
    (**cEnd)->SetOtherItemNode(*cStart);
  }
}    

bool FindColorChain (ListNode item, ListNode iEnd, ListNodeList &result) {
  bool bResult = false;

  // 1) find next color
  while (item != iEnd && (!(*item)->IsColor() || !(*item)->m_bStart)) {
    item++;
  }
  if (item == iEnd) {  // didn't find a color
    return false;
  }

  ItemType color1 = (*item)->m_ItemType;
  ListNode firstColor = item;
  result.push_back(item);
  
  while(1) {
    item = *((*item)->m_pOtherItemNode);  // color end tag
    result.push_back(item);

//    fprintf(stderr, "1\n");
//    (*item)->print(stderr);

    item++;
    
    while (item != iEnd && !(*item)->IsColor() && !(*item)->IsText() && !(*item)->IsWhitespace()) {
//      fprintf(stderr, "2\n");
//      (*item)->print(stderr);
      item++;
    }
    if (item == iEnd || (*item)->IsText() || (*item)->IsWhitespace() || ((*item)->IsColor() && !(*item)->m_bStart)) {
      break;
    }

    DocItem *pItem = (*item).Pointer();
//    fprintf(stderr, "3\n");
//    pItem->print(stderr);

    assert((*item)->IsColor() && (*item)->m_bStart);
    result.push_back(item);
  }    
  
  // start going backwards to find the same color end tag
  ListNodeList::iterator item2 = result.end();
  item2--;
  while (item2 != result.begin() && (*(*item2))->m_ItemType != color1) {
    item2--;
  }
  if (item2 == result.begin() || (*item2) == *((*firstColor)->m_pOtherItemNode)) {
    return false;
  }
  assert( (*(*item2))->m_bStart == false);
  item2++;  
  result.erase(item2, result.end());

//  fprintf(stderr, "found color chain:\n");
//  for (item2 = result.begin(); item2 != result.end(); item2++) {
//    (**item2)->print(stderr);
//  }

  return true;
}
