
#ifndef _Parser_h
#define _Parser_h

#include <stdio.h>
#include <string>
#include <map>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HOSTTYPE_FREEBSD
#include <hash_map>
  struct StringHash {
    size_t operator()(string const &s) const {
      return __stl_hash_string(s.c_str());
    }
  };
#endif

using namespace std;



class Parser {
public:
  typedef void (ParserCallback) (void *pArg, const char *pBuf, int nSize, int nLineno, bool bStart);

  Parser(ParserCallback *pTextCallback, ParserCallback *pWhitespaceCallback, ParserCallback *pDefaultTagCallback);
  void parse(const char *pSource, int nSize);
  void registerCallback(const char *pcTag, ParserCallback *pTagCallback, void *pArg);

  struct CallbackInfo {
    CallbackInfo() 
    : m_pCallback(NULL), m_pArg(NULL)
    {
    }
    CallbackInfo(ParserCallback *pCallback, void *pArg)
    : m_pCallback(pCallback), m_pArg(pArg)
    {
    }
    ParserCallback *m_pCallback;
    void *m_pArg;
  };

private:

  void HandleText (const char *pcSource, int nSize, int nLineno);
  void HandleWhitespace (const char *pcSource, int nSize, int nLineno);
  void HandleTag (const char *pcSource, int nSize, int nLineno);
  void ResizeBuf(const char *pcSource, int nSize);

  ParserCallback *m_pTextCallback, *m_pWhitespaceCallback, *m_pDefaultTagCallback;

#ifdef HOSTTYPE_FREEBSD
  typedef hash_map<string, CallbackInfo*, StringHash> CallbackMap;
#else
  typedef map<string, CallbackInfo*, less<string> > CallbackMap;
#endif

  CallbackMap m_CallbackMap;
  int m_nLineno;
  int m_nBufSize;
  char *m_pcBuf;
};

#endif //_Parser_h
