

#include <stdio.h>
#include <unistd.h>


#include "Parser.h"
#include "Globals.h"

int g_nText = 0;
int g_nWhitespace = 0;

#ifdef HOSTTYPE_FREEBSD
  #include <hash_map>
  typedef hash_map<string, int, StringHash> StringMapInt;
#else
  #include <map>
  typedef map<string, int, less<string> > StringMapInt;
#endif

StringMapInt g_StartTags, g_EndTags;

void HandleTag (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  StringMapInt::iterator item;
  StringMapInt *pMap = NULL;
  if (bStart) {
    pMap = &g_StartTags;
  } else {
    pMap = &g_EndTags; 
  }
  item = pMap->find(pcSource);
  if (item == pMap->end()) {
    (*pMap)[pcSource] = 1;
  } else {
    (*pMap)[pcSource]++;
  }
}

void HandleText (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  g_nText++;
}

void HandleWhitespace (void *pArg, const char *pcSource, int nSize, int nLineno, bool bStart) {
  g_nWhitespace++;
}

int main (int argc, char **argv) {
  Parser *pParser = new Parser(HandleText, HandleWhitespace, HandleTag);
  int size = 0, c = 0;
  char pcTemp[1024 * 1024 * 5];
 
  while ((c = read(STDIN_FILENO, pcTemp + size, 1024 * 1024 * 5)) > 0) {
    size += c;
  }
  pParser->parse(pcTemp, size);

  int nTotalItems = 0;
  printf ("%d text items\n", g_nText);
  nTotalItems += g_nText;
  printf ("%d whitespace items\n", g_nWhitespace);
  nTotalItems += g_nWhitespace;
  StringMapInt::iterator item, item2;
  int nTags;
  for (item = g_StartTags.begin(); item != g_StartTags.end(); item++) {
    item2 = g_EndTags.find((*item).first);
    nTags = 0;
    if (item2 != g_EndTags.end()) {
      nTags = (*item2).second;
      g_EndTags.erase(item2);
    }
    printf ("%d start tags, %d end tags for %s %s\n", (*item).second, nTags, (*item).first.c_str(), 
      (nTags == (*item).second) ? "" : " - mismatch");
    nTotalItems += nTags;
    nTotalItems += (*item).second;
  }
  for (item = g_EndTags.begin(); item != g_EndTags.end(); item++) {
    printf ("%d end tags for %s\n", (*item).second, (*item).first.c_str());
    nTotalItems += (*item).second;    
  }
  printf ("%d total items\n", nTotalItems);
  return 0;
}
