#!/home/ronnie/bin/tclsh8.3

#B
#(bold) set the B attribute
#
#EM
#(emphasis) invert the EM attribute
#
#I
#(italic) set the I attribute
#
#PL
#(plain) reset the U level to 0 and unset the B, EM, I, S,
# and TT attributes
#
#S
#(strong emphasis) set the S attribute
#
#TT
#(typewriter) set the TT attribute
#
#U
#(underline) increment the U level (but not above 3)
#
#0...9
#(size) set the size to 0...9
#
#r
#(color) set the color to red
#
#g
#set the color to green
#
#b
#set the color to blue
#
#c
#set the color to cyan
#
#m
#set the color to magenta
#
#y
#set the color to yellow
#
#k
#set the color to black
#
#w
#set the color to white

set tags {B EM I PL S TT U 0 1 2 3 4 5 6 7 8 9 r g b c m y k w}
set sizeTags {0 1 2 3 4 5 6 7 8 9};    # 0
set colorTags {r g b c m y k w};       # 1
set fontTags {B EM I PL S TT U};       # 2
array set tagTypes [list \
  0	$sizeTags	\
  1	$colorTags	\
  2	$fontTags	\
]

# 1) new tag, close tag, text, or whitespace
# 2) same tag type, or different
# 3) tag

# state variables
set tagStack {}
set lastItem 0;        # 0 - whitespace
                      # 1 - start tag
                      # 2 - end tag
                      # 3 - text
                      

array set nextItemProb {
  0,0	0
  0,1	40
  0,2	40
  0,3	20
  1,0	30	
  1,1	30
  1,2	10	
  1,3	30
  2,0	25
  2,1	25
  2,2	25
  2,3	25
  3,0	20
  3,1	40
  3,2	40
  3,3	0
}
  
proc ChooseNextItem {} {
  global tagStack lastItem tags sizeTags colorTags fontTags nextItemProb
  
  set rand [expr {rand()}]
  set range 0.0
  foreach pair [array names nextItemProb [set lastItem],*] {
    set nextItem [lindex [split $pair ,] end]
    set prob $nextItemProb($pair)
    #puts "$rand $range $prob"
    if {$rand >= $range && $rand < ($range + (($prob + 0.0) / 100))} {
      set lastItem $nextItem
      return $nextItem
    }
    set range [expr {$range + (($prob + 0.0) / 100)}]
  }
}

proc GenerateNextItem {} {
  global lastItem tagStack
  
  set found 0
  while {!$found} {
    set nextItem [ChooseNextItem]
    set found 1
    if {[llength $tagStack] == 0 && $nextItem == 2} {
      set found 0
    }
  }
  #puts $nextItem
  
  if {$nextItem == 0} {
    return [GenerateWhitespace]
  } elseif {$nextItem == 1} {
    return [GenerateStartTag]
  } elseif {$nextItem == 2} {
    return [GenerateEndTag] 
  } elseif {$nextItem == 3} {
    return [GenerateText]
  }
  
  return {}
}

proc GenerateWhitespace {} {
  return " \t\n "
}

proc GenerateText {} {
  return "asdf"
}

proc GetTagType {tag} {
  global tagTypes
  
  foreach {type list} [array get tagTypes] {
    if {[lsearch $list $tag] >= 0} {
      return $type
    }
  }
  
  return {}
}

proc GenerateStartTag {} {
  global tagStack tags sizeTags colorTags fontTags
  
  set lastTag [lindex $tagStack end]
  if {[string length $lastTag]} {;          # tag stack is not empty
    set tagType [GetTagType $lastTag]
    set rand [expr {rand()}]
    if {$rand < 0.4} {;                     # same tag type
      set nextTag [ChooseTag $tagType]
    } {
      set nextTag [ChooseTag {}];           # any tag type
    }
  } {
    set nextTag [ChooseTag {}]
  }
  
  lappend tagStack $nextTag
  return "<[set nextTag]>"
}

proc ChooseTag {tagType} {
  global tags tagTypes
  
  if {[string length $tagType]} {
    set tagList $tagTypes($tagType)
  } {
    set tagList $tags
  }
  
  set rand [expr {rand()}]
  set index [expr {int($rand * [llength $tagList])}]
  return [lindex $tagList $index]
}

proc GenerateEndTag {} {
  global tagStack
  
  set lastTag [lindex $tagStack end]
  set tagStack [lrange $tagStack 0 end-1]
  return "</[set lastTag]>"
}

proc CloseAllTags {} {
  global tagStack

  set result {}
  while {[llength $tagStack] > 0} {
    lappend result [GenerateEndTag]
  }
  return [join $result {}]
}

proc GenerateOutput {size} {
  
  set result {}
  set currentSize 0

  while {$currentSize < $size} {
    set item [GenerateNextItem]
    incr currentSize [string length $item]
#    lappend result $item
    puts -nonewline $item
  }

  set item [CloseAllTags]
#  lappend result $item
  puts -nonewline $item
  incr currentSize [string length $item]
#  puts [join $result {}]
}

set outputSize [lindex $argv 0]

fconfigure stdout -buffersize 1000000 -buffering full

expr srand([clock seconds])

GenerateOutput $outputSize

  
